/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * JBIListener.java - Initial Developper : ofabre
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.xslt.listener;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.logging.Level;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.security.auth.Subject;
import javax.xml.transform.Source;
import javax.xml.transform.Templates;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.exception.BusinessException;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.mbean.PerformanceNotification;
import org.ow2.petals.component.framework.su.ServiceUnitManager;
import org.ow2.petals.component.framework.util.ClassLoaderUtil;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.se.xslt.Constants;
import org.ow2.petals.se.xslt.StreamDataSource;
import org.ow2.petals.se.xslt.Xslt;
import org.ow2.petals.se.xslt.XsltImpl;

/**
 * The xslt binding component JBIListener
 * 
 * @author alouis, ofabre, cgirodengo
 * 
 */
public class JBIListener extends AbstractJBIListener {

    private static Map<String, Templates> TEMPLATES_CACHE = new Hashtable<String, Templates>();

    private Xslt xslt;

    /**
     * Transform the outputstream as an IN DataHandler
     * 
     * @param outputStream
     * @param attachmentName
     * @return
     */
    private DataHandler createDataSourceFromOutputStream(final ByteArrayOutputStream outputStream,
            final String attachmentName) {
        this.getLogger().log(Level.INFO, "Transform the result as an attachment.");
        final ByteArrayInputStream input = new ByteArrayInputStream(outputStream.toByteArray());
        final StreamDataSource datasource = new StreamDataSource(input, attachmentName);
        return new DataHandler(datasource);
    }

    /**
     * Transform the Outputstream as a Source
     * 
     * @param outputStream
     * @return
     */
    private StreamSource createSourceFromOutputStream(final ByteArrayOutputStream outputStream) {
        this.getLogger().log(Level.INFO, "Transform the result as Source.");

        final StreamSource sourceResult = new StreamSource();
        final ByteArrayInputStream result = new ByteArrayInputStream(outputStream.toByteArray());
        sourceResult.setInputStream(result);
        return sourceResult;
    }

    /**
     * Retrieve the Templates associated to this XSL, and update the
     * corresponding entry in the cache. Identifier is composed of a String
     * which value is suRootPath+xslName.
     * 
     * @param xslResourceInputStream
     * @param identifier
     * @param lastModified
     * @param canChange
     * @return
     * @throws BusinessException
     */
    private Templates getTemplatesAndUpdateCache(final InputStream xslResourceInputStream,
            final String identifier) throws BusinessException {
        Templates result = null;

        if (xslResourceInputStream == null) {
            return null;
        }

        final Source source = new StreamSource(xslResourceInputStream);

        try {
            final TransformerFactory transformerFactory = TransformerFactory.newInstance();
            result = transformerFactory.newTemplates(source);
        } catch (final TransformerConfigurationException e) {
            throw new BusinessException("Can't create XSL transformer", e);
        } catch (final TransformerFactoryConfigurationError e) {
            throw new BusinessException("Can't create XSL TransformerFactory", e);
        }

        JBIListener.TEMPLATES_CACHE.put(identifier, result);

        this.getLogger().log(Level.INFO, "XSL stored in cache. Identifier:" + identifier);
        return result;
    }

    /**
     * We retrieve the InputStream associated to the desired xsl from the su
     * directory
     * 
     * @param xslName
     * @param suRootPath
     * @return the xsl inputStream, NULL if the file does not exist
     */
    private InputStream getXslFileInputStream(final String xslName, final String suRootPath) {
        this.getLogger().fine(
                "Getting XSL sheet named " + xslName + " from directory " + suRootPath);
        FileInputStream fileIS = null;
        try {
            fileIS = new FileInputStream(suRootPath + File.separator + xslName);
        } catch (final FileNotFoundException e) {
            // nothing, there is no XSL file in the directory
        }
        return fileIS;
    }

    /**
     * We retrieve the InputStream associated to the desired xsl from the su
     * classloader
     * 
     * @param xslName
     * @param suRootPath
     * @return the xsl inputStream, NULL if the resource does not exist
     */
    private InputStream getXslResourceInputStream(final String xslName, final String suRootPath) {

        final ClassLoader classLoader = ClassLoaderUtil.createClassLoader(suRootPath, this
                .getClass().getClassLoader());

        return classLoader.getResourceAsStream(xslName);
    }

    /**
     * Retrieve the InputStream associated to the desired xsl from the SU
     * classloader or SU directory
     * 
     * @param serviceendpoint
     * @param extensions
     * @return
     */
    private Templates getXslTemplates(final ServiceEndpoint serviceendpoint,
            final Extensions extensions) throws BusinessException {
        this.getLogger().log(Level.INFO, "Retrieve the XSL style sheet to use.");

        final String suRootPath = ((ServiceUnitManager) this.getComponent().getServiceUnitManager())
                .getSUInstallRootForActivatedEp(serviceendpoint);
        final String xslName = extensions.getPetalsExtensions().getProperty(Constants.XSL_FILE);
        final String identifier = suRootPath + xslName;
        Templates xslTemplate = JBIListener.TEMPLATES_CACHE.get(identifier);

        if (xslTemplate == null) {
            this.getLogger().info("Getting XSD from " + suRootPath);

            // Try to retrieve the XSL from the classloader (XSL is in a JAR
            // file)
            InputStream xsl = this.getXslResourceInputStream(xslName, suRootPath);
            if (xsl == null) {
                // try to get the XSL from the filesystem
                xsl = this.getXslFileInputStream(xslName, suRootPath);
            }
            xslTemplate = this.getTemplatesAndUpdateCache(xsl, identifier);
        }
        return xslTemplate;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.framework.listener.AbstractListener#init()
     */
    @Override
    protected void init() {
        this.xslt = new XsltImpl();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.framework.listener.AbstractJBIListener#onJBIMessage(org.objectweb.petals.component.framework.util.Exchange)
     */
    @Override
    protected void onJBIMessage(final Exchange exchange) {
        this.getLogger().log(Level.INFO, "A new exchange is received.");

        try {
            if (exchange.isInOutPattern() || exchange.isInOptionalOutPattern()) {
                this.processExchange(exchange);
            } else {
                exchange.setError(new Exception("Exchange pattern (" + exchange.getPattern()
                        + ") not supported"));
            }
        } catch (final BusinessException e1) {
            try {
                exchange.setFault(e1);
            } catch (final MessagingException e) {
                this.getLogger().log(Level.SEVERE, "Can't return fault to consumer", e);
            }
        } catch (final MessagingException e) {
            exchange.setError(e);
        }
    }

    /**
     * Process the transformation and set the response as Source or attachment
     * 
     * @param exchange
     * @throws BusinessException
     * @throws MessagingException
     */
    private void processExchange(final Exchange exchange) throws BusinessException,
            MessagingException {
        this.getLogger().log(Level.INFO, "Process the transformation.");

        // Get in Message content
        final Source inContent = exchange.getInMessageContentAsSource();

        Subject subject = exchange.getInMessage().getSecuritySubject();

        final PerformanceNotification performanceNotification = this.component
                .createPerformanceNotificationSendingToOutside("Time for XSLT transformation",
                        exchange);

        final Extensions extensions = new Extensions(this.getProvides().getExtensions());

        // get the XSL template for current Service Unit
        final Templates xslTransformers = this.getXslTemplates(exchange.getEndpoint(), extensions);
        if (xslTransformers == null) {
            exchange.setError(new Exception(
                    "No XSL style sheet is found to transform the document."));
        }

        // transform the input
        final ByteArrayOutputStream outContent = this.xslt.transform(inContent, xslTransformers);

        final String outputAttachmentName = extensions.getPetalsExtensions().getProperty(
                Constants.XSL_OUTPUT_ATTACHMENT_NAME);
        if (StringHelper.isNullOrEmpty(outputAttachmentName)) {
            // set the response as output Source
            final Source source = this.createSourceFromOutputStream(outContent);
//            try {
//                System.out.println(SourceHelper.createString(inContent));
//                System.out.println(SourceHelper.createString(source));
//            } catch (PEtALSCDKException e) {
//                // TODO Auto-generated catch block
//                e.printStackTrace();
//            }
            exchange.setOutMessageContent(source, subject);
        } else {
            // set the response as attachment
            final DataHandler attachment = this.createDataSourceFromOutputStream(outContent,
                    outputAttachmentName);
            final HashMap<String, DataHandler> attachments = new HashMap<String, DataHandler>();
            attachments.put(outputAttachmentName, attachment);
            exchange.setOutMessageAttachements(attachments);
            final String outSource = "<attached-files><file-name>" + outputAttachmentName
                    + "</file-name></attached-files>";
            exchange.setOutMessageContent(outSource, subject);
        }

        this.component.sendPerformanceNotification(performanceNotification);
    }
}
